@@ -34,8 +34,16 @@
 
 package org.codehaus.janino;
 
-import java.io.*;
-import java.util.*;
+import java.io.DataOutputStream;
+import java.io.IOException;
+import java.util.ArrayList;
+import java.util.Arrays;
+import java.util.Collections;
+import java.util.HashMap;
+import java.util.Iterator;
+import java.util.List;
+import java.util.Map;
+import java.util.Stack;
 
 import org.codehaus.janino.util.ClassFile;
 
@@ -48,8 +56,7 @@ import org.codehaus.janino.util.ClassFile;
 public class CodeContext {
     private static final boolean DEBUG = false;
 
-    private static final int INITIAL_SIZE   = 100;
-    private static final int SIZE_INCREMENT = 128;
+    private static final int INITIAL_SIZE = 128;
 
     private /*final*/ ClassFile classFile;
 
@@ -135,11 +142,11 @@ public class CodeContext {
         DataOutputStream dos,
         short            lineNumberTableAttributeNameIndex
     ) throws IOException {
-        dos.writeShort(this.maxStack);                     // max_stack
-        dos.writeShort(this.maxLocals);                    // max_locals
-        dos.writeInt(0xffff & this.end.offset);            // code_length
-        dos.write(this.code, 0, 0xffff & this.end.offset); // code
-        dos.writeShort(this.exceptionTableEntries.size());        // exception_table_length
+        dos.writeShort(this.maxStack);                                // max_stack
+        dos.writeShort(this.maxLocals);                               // max_locals
+        dos.writeInt(this.end.offset);                                // code_length
+        dos.write(this.code, 0, this.end.offset);                     // code
+        dos.writeShort(this.exceptionTableEntries.size());            // exception_table_length
         for (int i = 0; i < this.exceptionTableEntries.size(); ++i) { // exception_table
             ExceptionTableEntry exceptionTableEntry = (ExceptionTableEntry) this.exceptionTableEntries.get(i);
             dos.writeShort(exceptionTableEntry.startPC.offset);
@@ -178,17 +185,21 @@ public class CodeContext {
      * Notice: On inconsistencies, a "RuntimeException" is thrown (KLUDGE).
      */
     public void flowAnalysis(String functionName) {
-        byte[] stackSizes = new byte[0xffff & this.end.offset];
+        if (CodeContext.DEBUG) { 
+            System.err.println("flowAnalysis(" + functionName + ")"); 
+        }
+        
+        byte[] stackSizes = new byte[this.end.offset];
         Arrays.fill(stackSizes, CodeContext.UNEXAMINED);
 
         // Analyze flow from offset zero.
         this.flowAnalysis(
             functionName,
-            this.code,                // code
-            0xffff & this.end.offset, // codeSize
-            0,                        // offset
-            0,                        // stackSize
-            stackSizes                // stackSizes
+            this.code,       // code
+            this.end.offset, // codeSize
+            0,               // offset
+            0,               // stackSize
+            stackSizes       // stackSizes
         );
 
         // Analyze flow from exception handler entry points.
@@ -196,14 +207,14 @@ public class CodeContext {
         while (analyzedExceptionHandlers != this.exceptionTableEntries.size()) {
             for (int i = 0; i < this.exceptionTableEntries.size(); ++i) {
                 ExceptionTableEntry exceptionTableEntry = (ExceptionTableEntry) this.exceptionTableEntries.get(i);
-                if (stackSizes[0xffff & exceptionTableEntry.startPC.offset] != CodeContext.UNEXAMINED) {
+                if (stackSizes[exceptionTableEntry.startPC.offset] != CodeContext.UNEXAMINED) {
                     this.flowAnalysis(
                         functionName,         
-                        this.code,                                                   // code
-                        0xffff & this.end.offset,                                    // codeSize
-                        0xffff & exceptionTableEntry.handlerPC.offset,               // offset
-                        stackSizes[0xffff & exceptionTableEntry.startPC.offset] + 1, // stackSize
-                        stackSizes                                                   // stackSizes
+                        this.code,                                          // code
+                        this.end.offset,                                    // codeSize
+                        exceptionTableEntry.handlerPC.offset,               // offset
+                        stackSizes[exceptionTableEntry.startPC.offset] + 1, // stackSize
+                        stackSizes                                          // stackSizes
                     );
                     ++analyzedExceptionHandlers;
                 }
@@ -287,7 +298,7 @@ public class CodeContext {
                 /* FALL THROUGH */
             case Opcode.SD_GETSTATIC:
                 stackSize += this.determineFieldSize((short) (
-                    (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
+                    extract16BitValue(0, operandOffset, code)
                 ));
                 break;
 
@@ -296,7 +307,7 @@ public class CodeContext {
                 /* FALL THROUGH */
             case Opcode.SD_PUTSTATIC:
                 stackSize -= this.determineFieldSize((short) (
-                    (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
+                    extract16BitValue(0, operandOffset, code)
                 ));
                 break;
 
@@ -307,7 +318,7 @@ public class CodeContext {
                 /* FALL THROUGH */
             case Opcode.SD_INVOKESTATIC:
                 stackSize -= this.determineArgumentsSize((short) (
-                    (code[operandOffset] << 8) + (0xff & code[operandOffset + 1])
+                    extract16BitValue(0, operandOffset, code)
                 ));
                 break;
 
@@ -368,13 +379,11 @@ public class CodeContext {
                 this.flowAnalysis(
                     functionName,
                     code, codeSize,
-                    offset + (
-                        ((       code[operandOffset++]) << 8) +
-                        ((0xff & code[operandOffset++])     )
-                    ),
+                    extract16BitValue(offset, operandOffset, code),
                     stackSize,
                     stackSizes
                 );
+                operandOffset += 2;
                 break;
 
             case Opcode.OP1_JSR:
@@ -384,10 +393,8 @@ public class CodeContext {
                     System.out.println(code[operandOffset]);
                     System.out.println(code[operandOffset + 1]);
                 }
-                int targetOffset = offset + (
-                    ((       code[operandOffset++]) << 8) +
-                    ((0xff & code[operandOffset++])     )
-                );
+                int targetOffset = extract16BitValue(offset, operandOffset, code);
+                operandOffset += 2;
                 if (stackSizes[targetOffset] == CodeContext.UNEXAMINED) {
                     this.flowAnalysis(
                         functionName,
@@ -403,14 +410,10 @@ public class CodeContext {
                 this.flowAnalysis(
                     functionName,
                     code, codeSize,
-                    offset + (
-                        ((       code[operandOffset++]) << 24) +
-                        ((0xff & code[operandOffset++]) << 16) +
-                        ((0xff & code[operandOffset++]) <<  8) +
-                        ((0xff & code[operandOffset++])      )
-                    ),
+                    extract32BitValue(offset, operandOffset, code),
                     stackSize, stackSizes
                 );
+                operandOffset += 4;
                 break;
 
             case Opcode.OP1_LOOKUPSWITCH:
@@ -418,33 +421,23 @@ public class CodeContext {
                 this.flowAnalysis(
                     functionName,
                     code, codeSize,
-                    offset + (
-                        ((       code[operandOffset++]) << 24) +
-                        ((0xff & code[operandOffset++]) << 16) +
-                        ((0xff & code[operandOffset++]) <<  8) +
-                        ((0xff & code[operandOffset++])      )
-                    ),
+                    extract32BitValue(offset, operandOffset, code),
                     stackSize, stackSizes
                 );
-                int npairs = (
-                    ((       code[operandOffset++]) << 24) +
-                    ((0xff & code[operandOffset++]) << 16) +
-                    ((0xff & code[operandOffset++]) <<  8) +
-                    ((0xff & code[operandOffset++])      )
-                );
+                operandOffset += 4;
+                
+                int npairs = extract32BitValue(0, operandOffset, code);
+                operandOffset += 4;
+                
                 for (int i = 0; i < npairs; ++i) {
-                    operandOffset += 4;
+                    operandOffset += 4; //skip match value
                     this.flowAnalysis(
                         functionName,
                         code, codeSize,
-                        offset + (
-                            ((       code[operandOffset++]) << 24) +
-                            ((0xff & code[operandOffset++]) << 16) +
-                            ((0xff & code[operandOffset++]) <<  8) +
-                            ((0xff & code[operandOffset++])      )
-                        ),
+                        extract32BitValue(offset, operandOffset, code),
                         stackSize, stackSizes
                     );
+                    operandOffset += 4; //advance over offset
                 }
                 break;
 
@@ -453,38 +446,22 @@ public class CodeContext {
                 this.flowAnalysis(
                     functionName,
                     code, codeSize,
-                    offset + (
-                        ((       code[operandOffset++]) << 24) +
-                        ((0xff & code[operandOffset++]) << 16) +
-                        ((0xff & code[operandOffset++]) <<  8) +
-                        ((0xff & code[operandOffset++])      )
-                    ),
+                    extract32BitValue(offset, operandOffset, code),
                     stackSize, stackSizes
                 );
-                int low = (
-                    ((       code[operandOffset++]) << 24) +
-                    ((0xff & code[operandOffset++]) << 16) +
-                    ((0xff & code[operandOffset++]) <<  8) +
-                    ((0xff & code[operandOffset++])      )
-                );
-                int hi = (
-                    ((       code[operandOffset++]) << 24) +
-                    ((0xff & code[operandOffset++]) << 16) +
-                    ((0xff & code[operandOffset++]) <<  8) +
-                    ((0xff & code[operandOffset++])      )
-                );
+                operandOffset += 4;
+                int low = extract32BitValue(offset, operandOffset, code);
+                operandOffset += 4;
+                int hi = extract32BitValue(offset, operandOffset, code);
+                operandOffset += 4;
                 for (int i = low; i <= hi; ++i) {
                     this.flowAnalysis(
                         functionName,
                         code, codeSize,
-                        offset + (
-                            ((       code[operandOffset++]) << 24) +
-                            ((0xff & code[operandOffset++]) << 16) +
-                            ((0xff & code[operandOffset++]) <<  8) +
-                            ((0xff & code[operandOffset++])      )
-                        ),
+                        extract32BitValue(offset, operandOffset, code),
                         stackSize, stackSizes
                     );
+                    operandOffset += 4;
                 }
                 break;
 
@@ -530,20 +507,92 @@ public class CodeContext {
             offset = operandOffset;
         }
     }
+    
+    /**
+     * Extract a 16 bit value at offset in code and add bias to it
+     * @param bias    an int to skew the final result by (useful for calculating relative offsets)
+     * @param offset  the position in the code array to extract the bytes from
+     * @param code    the array of bytes
+     * @return an integer that treats the two bytes at position offset as an UNSIGNED SHORT
+     */
+    private int extract16BitValue(int bias, int offset, byte[] code) {
+        int res = bias + (
+                ((       code[offset  ]) << 8) +
+                ((0xff & code[offset+1])     )
+        );
+        if (CodeContext.DEBUG) {
+            System.out.println("extract16BitValue(bias, offset) = (" + bias +", " + offset + ")");
+            System.out.println("bytes = {" + code[offset] + ", " + code[offset+1] + "}");
+            System.out.println("result = " + res);
+        }
+        return res;
+    }
+    
+    /**
+     * Extract a 32 bit value at offset in code and add bias to it
+     * @param bias    an int to skew the final result by (useful for calculating relative offsets)
+     * @param offset  the position in the code array to extract the bytes from
+     * @param code    the array of bytes
+     * @return the 4 bytes at position offset + bias
+     */
+    private int extract32BitValue(int bias, int offset, byte[] code) {
+        int res = bias + (
+                ((       code[offset  ]) << 24) +
+                ((0xff & code[offset+1]) << 16) +
+                ((0xff & code[offset+2]) <<  8) +
+                ((0xff & code[offset+3])      )
+        );
+        if (CodeContext.DEBUG) {
+            System.out.println("extract32BitValue(bias, offset) = (" + bias +", " + offset + ")");
+            System.out.println(
+                    "bytes = {" + 
+                    code[offset  ] + ", " + 
+                    code[offset+1] + ", " + 
+                    code[offset+2] + ", " + 
+                    code[offset+3] + "}"
+            );
+            System.out.println("result = " + res);
+        }
+        return res;
+    }
+    
+    /**
+     * fixUp() all of the offsets and relocate() all relocatables
+     */
+    public void fixUpAndRelocate() {
+
+        // We do this in a loop to allow relocatables to adjust the size
+        // of things in the byte stream.  It is extremely unlikely, but possible
+        // that a late relocatable will grow the size of the bytecode, and require
+        // an earlier relocatable to switch from 32K mode to 64K mode branching
+        do {
+            fixUp();
+        } while (!relocate());
+    }
 
     /**
      * Fix up all offsets.
      */
-    public void fixUp() {
+    private void fixUp() {
         for (Offset o = this.beginning; o != this.end; o = o.next) {
             if (o instanceof FixUp) ((FixUp) o).fixUp();
         }
     }
 
-    public void relocate() {
+    /**
+     * Relocate all relocatables and aggregate their response into a single one
+     * @return true if all of them relocated successfully
+     *         false if any of them needed to change size
+     */
+    private boolean relocate() {
+        boolean finished = true;
         for (int i = 0; i < this.relocatables.size(); ++i) {
-            ((Relocatable) this.relocatables.get(i)).relocate();
+            //do not terminate earlier so that everything gets a chance to grow in the first pass
+            //changes the common case for this to be O(n) instead of O(n**2)
+            boolean part =  ((Relocatable) this.relocatables.get(i)).relocate();
+            finished = finished && part;
         }
+        return finished;
     }
 
     /**
@@ -610,6 +659,97 @@ public class CodeContext {
      */
     public void write(short lineNumber, byte[] b) {
         if (b.length == 0) return;
+        
+        int ico = this.currentInserter.offset;
+        this.makeSpace(lineNumber, b.length);
+        System.arraycopy(b, 0, this.code, ico, b.length);
+    }
+
+    /**
+     * Inserts a byte at the current insertion position. Creates
+     * {@link LineNumberOffset}s as necessary.
+     * <p>
+     * This method is an optimization to avoid allocating small byte[] and ease
+     * GC load.
+     * 
+     * @param lineNumber The line number that corresponds to the byte code, or -1
+     * @param b1
+     */
+    public void write(short lineNumber, byte b1) {
+        int ico = this.currentInserter.offset;
+        this.makeSpace(lineNumber, 1);
+        this.code[ico] = b1;
+    }
+    
+    /**
+     * Inserts bytes at the current insertion position. Creates
+     * {@link LineNumberOffset}s as necessary.
+     * <p>
+     * This method is an optimization to avoid allocating small byte[] and ease
+     * GC load.
+     * 
+     * @param lineNumber The line number that corresponds to the byte code, or -1
+     * @param b1
+     * @param b2
+     */
+    public void write(short lineNumber, byte b1, byte b2) {
+        int ico = this.currentInserter.offset;
+        this.makeSpace(lineNumber, 2);
+        this.code[ico++] = b1;
+        this.code[ico  ] = b2;
+    }
+    
+    /**
+     * Inserts bytes at the current insertion position. Creates
+     * {@link LineNumberOffset}s as necessary.
+     * <p>
+     * This method is an optimization to avoid allocating small byte[] and ease
+     * GC load.
+     * 
+     * @param lineNumber The line number that corresponds to the byte code, or -1
+     * @param b1
+     * @param b2
+     * @param b3
+     */
+    public void write(short lineNumber, byte b1, byte b2, byte b3) {
+        int ico = this.currentInserter.offset;
+        this.makeSpace(lineNumber, 3);
+        this.code[ico++] = b1;
+        this.code[ico++] = b2;
+        this.code[ico  ] = b3;
+    }
+    
+    /**
+     * Inserts bytes at the current insertion position. Creates
+     * {@link LineNumberOffset}s as necessary.
+     * <p>
+     * This method is an optimization to avoid allocating small byte[] and ease
+     * GC load.
+     * 
+     * @param lineNumber The line number that corresponds to the byte code, or -1
+     * @param b1
+     * @param b2
+     * @param b3
+     * @param b4
+     */
+    public void write(short lineNumber, byte b1, byte b2, byte b3, byte b4) {
+        int ico = this.currentInserter.offset;
+        this.makeSpace(lineNumber, 4);
+        this.code[ico++] = b1;
+        this.code[ico++] = b2;
+        this.code[ico++] = b3;
+        this.code[ico  ] = b4;
+    }
+    
+    /**
+     * Add space for size bytes at current offset. Creates
+     * {@link LineNumberOffset}s as necessary.
+     *  
+     * @param lineNumber The line number that corresponds to the byte code, or -1
+     * @param size       The size in bytes to inject
+     */
+    public void makeSpace(short lineNumber, int size) {
+        if (size == 0) return;
 
         INSERT_LINE_NUMBER_OFFSET:
         if (lineNumber != -1) {
@@ -627,54 +767,151 @@ public class CodeContext {
             this.currentInserter.prev = lno;
         }
 
-        int ico = 0xffff & this.currentInserter.offset;
-        if ((0xffff & this.end.offset) + b.length <= this.code.length) {
-            System.arraycopy(this.code, ico, this.code, ico + b.length, (0xffff & this.end.offset) - ico);
+        int ico = this.currentInserter.offset;
+        if (this.end.offset + size <= this.code.length) {
+            // Optimization to avoid a trivial method call in the common case
+            if(ico != this.end.offset) {
+                System.arraycopy(this.code, ico, this.code, ico + size, this.end.offset - ico);
+            }
         } else {
             byte[] oldCode = this.code;
-            this.code = new byte[this.code.length + CodeContext.SIZE_INCREMENT];
-            if (this.code.length >= 0xffff) throw new RuntimeException("Code attribute in class \"" + this.classFile.getThisClassName() + "\" grows beyond 64 KB");
+            //double size to avoid horrible performance, but don't grow over our limit
+            int newSize = Math.max(Math.min(oldCode.length * 2, 0xffff), oldCode.length + size);
+            if (newSize > 0xffff) throw new RuntimeException("Code attribute in class \"" + this.classFile.getThisClassName() + "\" grows beyond 64 KB");
+            this.code = new byte[newSize];
             System.arraycopy(oldCode, 0, this.code, 0, ico);
-            System.arraycopy(oldCode, ico, this.code, ico + b.length, (0xffff & this.end.offset) - ico);
-        }
-        System.arraycopy(b, 0, this.code, ico, b.length);
-        for (Offset o = this.currentInserter; o != null; o = o.next) o.offset += b.length;
+            System.arraycopy(oldCode, ico, this.code, ico + size, this.end.offset - ico);
+        }  
+        Arrays.fill(this.code, ico, ico + size, (byte)0);
+        for (Offset o = this.currentInserter; o != null; o = o.next) o.offset += size;
     }
 
     /**
      * @param lineNumber The line number that corresponds to the byte code, or -1
      */
     public void writeShort(short lineNumber, int v) {
-        this.write(lineNumber, new byte[] { (byte) (v >> 8), (byte) v });
+        this.write(lineNumber, (byte) (v >> 8), (byte) v);
     }
 
     /**
      * @param lineNumber The line number that corresponds to the byte code, or -1
      */
     public void writeBranch(short lineNumber, int opcode, final Offset dst) {
-        this.relocatables.add(new Branch(this.newOffset(), dst));
-        this.write(lineNumber, new byte[] { (byte) opcode, -1, -1 });
+        this.relocatables.add(new Branch(opcode, dst));
+        this.write(lineNumber, (byte) opcode, (byte) -1, (byte) -1);
     }
 
     private class Branch extends Relocatable {
-        public Branch(Offset source, Offset destination) {
-            this.source = source;
+        public Branch(int opcode, Offset destination) {
+            this.opcode = opcode;
+            this.source = CodeContext.this.newInserter();
             this.destination = destination;
+            if (opcode == Opcode.JSR_W || opcode == Opcode.GOTO_W) {
+                //no need to expand wide opcodes
+                this.expanded = true;
+            } else { 
+                this.expanded = false;
+            }
         }
-        public void relocate() {
+        
+        public boolean relocate() {
             if (this.destination.offset == Offset.UNSET) throw new RuntimeException("Cannot relocate branch to unset destination offset");
             int offset = this.destination.offset - this.source.offset;
-            if (offset > Short.MAX_VALUE || offset < Short.MIN_VALUE) throw new RuntimeException("Branch offset out of range");
-            byte[] ba = new byte[] { (byte) (offset >> 8), (byte) offset };
-            System.arraycopy(ba, 0, CodeContext.this.code, (0xffff & this.source.offset) + 1, 2);
+            
+            if (!this.expanded && (offset > Short.MAX_VALUE || offset < Short.MIN_VALUE)) {
+                //we want to insert the data without skewing our source position,
+                //so we will cache it and then restore it later.
+                int pos = this.source.offset; 
+                CodeContext.this.pushInserter(this.source); {
+                    // promotion to a wide instruction only requires 2 extra bytes 
+                    // everything else requires a new GOTO_W instruction after a negated if
+                    CodeContext.this.makeSpace(
+                        (short) -1, 
+                        this.opcode == Opcode.GOTO ? 2 : this.opcode == Opcode.JSR ? 2 : 5
+                    );
+                } CodeContext.this.popInserter();
+                this.source.offset = pos;
+                this.expanded = true;
+                return false;
+            }
+            
+            final byte[] ba;
+            if (!this.expanded) {
+                //we fit in a 16-bit jump
+                ba = new byte[] { (byte) this.opcode, (byte) (offset >> 8), (byte) offset };
+            } else {
+                if (this.opcode == Opcode.GOTO || this.opcode == Opcode.JSR) {
+                    ba = new byte[] { 
+                        (byte) (this.opcode + 33), // GOTO => GOTO_W; JSR => JSR_W
+                        (byte) (offset >> 24), 
+                        (byte) (offset >> 16), 
+                        (byte) (offset >> 8), 
+                        (byte) offset 
+                    };
+                } else
+                {
+                    //exclude the if-statement from jump target
+                    //if jumping backwards this will increase the jump to go over it
+                    //if jumping forwards this will decrease the jump by it
+                    offset -= 3;
+                            
+                    //  [if cond offset]
+                    //expands to 
+                    //  [if !cond skip_goto]
+                    //  [GOTO_W offset]
+                    ba = new byte[] { 
+                        CodeContext.invertBranchOpcode((byte) this.opcode),
+                        (byte) 0,
+                        (byte) 8, //jump from this instruction past the GOTO_W
+                        (byte) Opcode.GOTO_W, 
+                        (byte) (offset >> 24), 
+                        (byte) (offset >> 16), 
+                        (byte) (offset >> 8), 
+                        (byte) offset 
+                    };
+                }
+            }
+            System.arraycopy(ba, 0, CodeContext.this.code, this.source.offset, ba.length);
+            return true;
         }
-        private final Offset source;
+        
+        private boolean expanded; //marks whether this has been expanded to account for a wide branch
+        private final int opcode;
+        private final Inserter source;
         private final Offset destination;
     }
 
+    /**
+     * E.g. {@link Opcode#IFLT} ("less than") inverts to {@link Opcode#IFGE} ("greater than or equal to").
+     */
+    private static byte invertBranchOpcode(byte branchOpcode) {
+        return ((Byte) CodeContext.BRANCH_OPCODE_INVERSION.get(new Byte(branchOpcode))).byteValue();
+    }
+    private static final Map BRANCH_OPCODE_INVERSION = CodeContext.createBranchOpcodeInversion(); // Map<Byte branch-opcode, Byte inverted-branch-opcode>
+    private static Map createBranchOpcodeInversion() {
+        Map m = new HashMap();
+        m.put(new Byte(Opcode.IF_ACMPEQ), new Byte(Opcode.IF_ACMPNE)); 
+        m.put(new Byte(Opcode.IF_ACMPNE), new Byte(Opcode.IF_ACMPEQ)); 
+        m.put(new Byte(Opcode.IF_ICMPEQ), new Byte(Opcode.IF_ICMPNE)); 
+        m.put(new Byte(Opcode.IF_ICMPNE), new Byte(Opcode.IF_ICMPEQ)); 
+        m.put(new Byte(Opcode.IF_ICMPGE), new Byte(Opcode.IF_ICMPLT)); 
+        m.put(new Byte(Opcode.IF_ICMPLT), new Byte(Opcode.IF_ICMPGE)); 
+        m.put(new Byte(Opcode.IF_ICMPGT), new Byte(Opcode.IF_ICMPLE)); 
+        m.put(new Byte(Opcode.IF_ICMPLE), new Byte(Opcode.IF_ICMPGT)); 
+        m.put(new Byte(Opcode.IFEQ),      new Byte(Opcode.IFNE)); 
+        m.put(new Byte(Opcode.IFNE),      new Byte(Opcode.IFEQ)); 
+        m.put(new Byte(Opcode.IFGE),      new Byte(Opcode.IFLT)); 
+        m.put(new Byte(Opcode.IFLT),      new Byte(Opcode.IFGE)); 
+        m.put(new Byte(Opcode.IFGT),      new Byte(Opcode.IFLE)); 
+        m.put(new Byte(Opcode.IFLE),      new Byte(Opcode.IFGT)); 
+        m.put(new Byte(Opcode.IFNULL),    new Byte(Opcode.IFNONNULL)); 
+        m.put(new Byte(Opcode.IFNONNULL), new Byte(Opcode.IFNULL));
+        return Collections.unmodifiableMap(m);
+    }
+
     public void writeOffset(short lineNumber, Offset src, final Offset dst) {
         this.relocatables.add(new OffsetBranch(this.newOffset(), src, dst));
-        this.write(lineNumber, new byte[] { -1, -1, -1, -1 });
+        this.write(lineNumber, (byte) -1, (byte) -1, (byte) -1, (byte) -1);
     }
 
     private class OffsetBranch extends Relocatable {
@@ -683,7 +920,7 @@ public class CodeContext {
             this.source      = source;
             this.destination = destination;
         }
-        public void relocate() {
+        public boolean relocate() {
             if (
                 this.source.offset == Offset.UNSET ||
                 this.destination.offset == Offset.UNSET
@@ -695,7 +932,8 @@ public class CodeContext {
                 (byte) (offset >> 8),
                 (byte) offset
             };
-            System.arraycopy(ba, 0, CodeContext.this.code, 0xffff & this.where.offset, 4);
+            System.arraycopy(ba, 0, CodeContext.this.code, this.where.offset, 4);
+            return true;
         }
         private final Offset where, source, destination;
     }
@@ -755,9 +993,9 @@ public class CodeContext {
      * automatically shifted.
      */
     public class Offset {
-        short              offset = Offset.UNSET;
+        int              offset = Offset.UNSET;
         Offset             prev = null, next = null;
-        final static short UNSET = -1;
+        final static int UNSET = -1;
 
         /**
          * Set this "Offset" to the offset of the current inserter; insert
@@ -776,7 +1014,7 @@ public class CodeContext {
         public final CodeContext getCodeContext() { return CodeContext.this; }
 
         public String toString() {
-            return CodeContext.this.classFile.getThisClassName() + ": " + (0xffff & this.offset);
+            return CodeContext.this.classFile.getThisClassName() + ": " + this.offset;
         }
     }
 
@@ -830,15 +1068,20 @@ public class CodeContext {
     }
 
     public class LineNumberOffset extends Offset {
-        private final short lineNumber;
-        public LineNumberOffset(short offset, short lineNumber) {
+        private final int lineNumber;
+        public LineNumberOffset(int offset, int lineNumber) {
             this.lineNumber = lineNumber;
             this.offset = offset;
         }
     }
 
     private abstract class Relocatable {
-        public abstract void relocate();
+        /**
+         * Relocate this object.
+         * @return true if the relocation succeeded in place
+         *         false if the relocation grew the number of bytes required
+         */
+        public abstract boolean relocate();
     }
 
     private short       localVariableArrayLength = 0;
