@@ -375,13 +375,13 @@ public class Scanner {
     }
 
     /**
-     * This value represents the "magic" literal "2147483648" which is only
+     * This reference represents the "magic" literal "2147483648" which is only
      * allowed in a negated context.
      */
     public static final Integer MAGIC_INTEGER = new Integer(Integer.MIN_VALUE);
 
     /**
-     * This value represents the "magic" literal "9223372036854775808L" which is only
+     * This reference represents the "magic" literal "9223372036854775808L" which is only
      * allowed in a negated context.
      */
     public static final Long MAGIC_LONG = new Long(Long.MIN_VALUE);
@@ -404,6 +404,9 @@ public class Scanner {
     public final class LiteralToken extends Token {
         private final Object value;
 
+        /**
+         * @param value A {@link Boolean}, {@link String}, {@link Double}, {@link Float}, {@link Character}, or <code>null</code>
+         */
         public LiteralToken(Object value) {
             this.value = value;
         }
@@ -442,11 +445,13 @@ public class Scanner {
         }
         if (v instanceof Integer) {
             if (v == Scanner.MAGIC_INTEGER) return "2147483648";
-            return v.toString();
+            int iv = ((Integer) v).intValue();
+            return iv < 0 ? "0x" + Integer.toHexString(iv) : Integer.toString(iv);
         }
         if (v instanceof Long) {
             if (v == Scanner.MAGIC_LONG) return "9223372036854775808L";
-            return v.toString() + 'L';
+            long lv = ((Long) v).longValue();
+            return lv < 0L ? "0x" + Long.toHexString(lv) + 'L' : Long.toString(lv) + 'L';
         }
         if (v instanceof Float) {
             return v.toString() + 'F';
@@ -1106,6 +1111,9 @@ public class Scanner {
         return new LiteralToken(new Double(d));
     }
 
+    /**
+     * Consume characters until a literal character is complete.
+     */
     private char unescapeCharacterLiteral() throws ScanException, IOException {
         if (this.nextChar == -1) throw new ScanException("EOF in character literal");
 
