@@ -201,7 +201,7 @@ public class SimpleCompiler extends Cookable {
     public static final ClassLoader BOOT_CLASS_LOADER = new ClassLoader(null) {};
 
     /**
-     * Allowe references to the classes loaded through this parent class loader
+     * Allow references to the classes loaded through this parent class loader
      * (@see {@link #setParentClassLoader(ClassLoader)}), plus the extra
      * <code>auxiliaryClasses</code>.
      * <p>
@@ -211,6 +211,7 @@ public class SimpleCompiler extends Cookable {
      * <code>optionalParentClassLoader</code>.
      */
     public void setParentClassLoader(ClassLoader optionalParentClassLoader, Class[] auxiliaryClasses) {
+        assertNotCooked();
         this.parentClassLoader = (
             optionalParentClassLoader != null
             ? optionalParentClassLoader
@@ -232,6 +233,20 @@ public class SimpleCompiler extends Cookable {
             DebuggingInformation.DEFAULT_DEBUGGING_INFORMATION
         );
     }
+    
+    /**
+     * Cook this compilation unit directly. 
+     *  See {@link Cookable#cook}
+     */
+    public void cook(Java.CompilationUnit compilationUnit) throws CompileException {
+        this.setUpClassLoaders();
+
+        // Compile the classes and load them.
+        this.compileToClassLoader(
+            compilationUnit,
+            DebuggingInformation.DEFAULT_DEBUGGING_INFORMATION
+        );
+    }
 
     /**
      * Initializes {@link #classLoader} and {@link #iClassLoader} from the configured
@@ -239,7 +254,7 @@ public class SimpleCompiler extends Cookable {
      * {@link #classToType(Location, Class)} and friends which are used when creating the AST.
      */
     protected final void setUpClassLoaders() {
-        if (this.classLoader != null) throw new RuntimeException("\"cook()\" must not be called more than once");
+        assertNotCooked();
 
         // Set up the ClassLoader for the compilation and the loading.
         this.classLoader = new AuxiliaryClassLoader(this.parentClassLoader);
@@ -437,4 +452,11 @@ public class SimpleCompiler extends Cookable {
         );
         return this.result;
     }
+
+    /**
+     * Throw an {@link IllegalStateException} if this {@link Cookable} is already cooked.
+     */
+    protected void assertNotCooked() {
+        if (this.classLoader != null) throw new IllegalStateException("Already cooked");
+    }
 }
