@@ -44,7 +44,7 @@ import java.util.*;
  * This class traverses the subnodes of an AST. Derived classes may override
  * individual methods to process specific nodes, e.g.:<pre>
  *     LocalClassDeclaration lcd = ...;
- *     lcd.accept(new Traverser {
+ *     lcd.accept(new Traverser() {
  *         int n = 0;
  *         public void traverseMethodDeclarator(Java.MethodDeclarator md) {
  *             ++this.n;
@@ -93,7 +93,6 @@ public class Traverser {
         public final void visitCast(Java.Cast c)                                                              { Traverser.this.traverseCast(c); }
         public final void visitClassLiteral(Java.ClassLiteral cl)                                             { Traverser.this.traverseClassLiteral(cl); }
         public final void visitConditionalExpression(Java.ConditionalExpression ce)                           { Traverser.this.traverseConditionalExpression(ce); }
-        public final void visitConstantValue(Java.ConstantValue cv)                                           { Traverser.this.traverseConstantValue(cv); }
         public final void visitCrement(Java.Crement c)                                                        { Traverser.this.traverseCrement(c); }
         public final void visitInstanceof(Java.Instanceof io)                                                 { Traverser.this.traverseInstanceof(io); }
         public final void visitMethodInvocation(Java.MethodInvocation mi)                                     { Traverser.this.traverseMethodInvocation(mi); }
@@ -365,10 +364,6 @@ public class Traverser {
         this.traverseRvalue(ce);
     }
 
-    public void traverseConstantValue(Java.ConstantValue cv) {
-        this.traverseRvalue(cv);
-    }
-
     public void traverseCrement(Java.Crement c) {
         c.operand.accept((Visitor.LvalueVisitor) this.cv);
         this.traverseRvalue(c);
